'''ResNet in PyTorch.

For Pre-activation ResNet, see 'preact_resnet.py'.

Reference:
[1] Kaiming He, Xiangyu Zhang, Shaoqing Ren, Jian Sun
    Deep Residual Learning for Image Recognition. arXiv:1512.03385
'''
import torch.nn as nn
import torch.nn.functional as F


class BasicBlock(nn.Module):
    expansion = 1

    def __init__(self, in_planes, planes, stride=1):
        super(BasicBlock, self).__init__()
        self.conv1 = nn.Conv2d(
            in_planes, planes, kernel_size=3, stride=stride, padding=1, bias=False)
        self.bn1 = nn.BatchNorm2d(planes)
        self.conv2 = nn.Conv2d(planes, planes, kernel_size=3,
                               stride=1, padding=1, bias=False)
        self.bn2 = nn.BatchNorm2d(planes)

        self.shortcut = nn.Sequential()
        if stride != 1 or in_planes != self.expansion*planes:
            self.shortcut = nn.Sequential(
                nn.Conv2d(in_planes, self.expansion*planes,
                          kernel_size=1, stride=stride, bias=False),
                nn.BatchNorm2d(self.expansion*planes)
            )

    def forward(self, x):
        out = F.relu(self.bn1(self.conv1(x)))
        out = self.bn2(self.conv2(out))
        out += self.shortcut(x)
        out = F.relu(out)
        return out


class Bottleneck(nn.Module):
    expansion = 4

    def __init__(self, in_planes, planes, stride=1):
        super(Bottleneck, self).__init__()
        self.conv1 = nn.Conv2d(in_planes, planes, kernel_size=1, bias=False)
        self.bn1 = nn.BatchNorm2d(planes)
        self.conv2 = nn.Conv2d(planes, planes, kernel_size=3,
                               stride=stride, padding=1, bias=False)
        self.bn2 = nn.BatchNorm2d(planes)
        self.conv3 = nn.Conv2d(planes, self.expansion *
                               planes, kernel_size=1, bias=False)
        self.bn3 = nn.BatchNorm2d(self.expansion*planes)

        self.shortcut = nn.Sequential()
        if stride != 1 or in_planes != self.expansion*planes:
            self.shortcut = nn.Sequential(
                nn.Conv2d(in_planes, self.expansion*planes,
                          kernel_size=1, stride=stride, bias=False),
                nn.BatchNorm2d(self.expansion*planes)
            )

    def forward(self, x):
        out = F.relu(self.bn1(self.conv1(x)))
        out = F.relu(self.bn2(self.conv2(out)))
        out = self.bn3(self.conv3(out))
        out += self.shortcut(x)
        out = F.relu(out)
        return out


class ResNetCon(nn.Module):
    def __init__(self, block, num_blocks, num_classes=10, rotation=True, rotnet=None, classifier_bias=True,loss_fn='con'):
        super(ResNetCon, self).__init__()
        self.in_planes = 64

        self.conv1 = nn.Conv2d(3, 64, kernel_size=3,
                               stride=1, padding=1, bias=False)
        self.bn1 = nn.BatchNorm2d(64)
        self.layer1 = self._make_layer(block, 64, num_blocks[0], stride=1)
        self.layer2 = self._make_layer(block, 128, num_blocks[1], stride=2)
        self.layer3 = self._make_layer(block, 256, num_blocks[2], stride=2)
        self.layer4 = self._make_layer(block, 512, num_blocks[3], stride=2)
        # self.output = nn.Linear(512*block.expansion, num_classes, bias=classifier_bias)
        self.channels = 512 * block.expansion
        self.fc = nn.Linear(512*block.expansion, num_classes, bias=classifier_bias)
        self.fc1 = nn.Linear(512*block.expansion, num_classes, bias=classifier_bias)
        channels = self.channels
        feat_dim = 2 * channels
        self.head = nn.Sequential(nn.Linear(channels, channels), nn.BatchNorm1d(channels),
                                  nn.ReLU(inplace=True),
                                  nn.Linear(channels, feat_dim))

        self.head_fc = nn.Sequential(nn.Linear(channels, channels), nn.BatchNorm1d(channels),
                                     nn.ReLU(inplace=True),
                                     nn.Linear(channels, feat_dim))

        self.rotation = rotation
        if self.rotation:
            if rotnet is not None:
                self.rot = rotnet
            else:
                self.rot = nn.Linear(512*block.expansion, 4)

        # Initialize the weights
        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                nn.init.kaiming_normal_(m.weight, mode='fan_out', nonlinearity='relu')
            elif isinstance(m, nn.BatchNorm2d):
                nn.init.constant_(m.weight, 1)
                nn.init.constant_(m.bias, 0)
            elif isinstance(m, nn.Linear):
                nn.init.xavier_normal_(m.weight)
                if m.bias is not None:
                    nn.init.constant_(m.bias, 0)

    def _make_layer(self, block, planes, num_blocks, stride):
        strides = [stride] + [1]*(num_blocks-1)
        layers = []
        for stride in strides:
            layers.append(block(self.in_planes, planes, stride))
            self.in_planes = planes * block.expansion
        return nn.Sequential(*layers)

    def forward(self, x, classify=False, classify1=False):
        assert int(classify) + int(classify1) <= 1
        if classify == True:
            return self.fc(x)
        elif classify1 == True:
            return self.fc1(x)

        out = F.relu(self.bn1(self.conv1(x)))
        out = self.layer1(out)
        out = self.layer2(out)
        out = self.layer3(out)
        out = self.layer4(out)
        f = F.adaptive_avg_pool2d(out, 1)
        # c = self.output(f.squeeze())
        out = f.view(-1, self.channels)
        if self.training:
            feat_mlp = F.normalize(self.head(out), dim=1)
            centers_logits = F.normalize(self.head_fc(self.fc1.weight), dim=1)
            return out, feat_mlp, centers_logits
        else:
            return out
        return out

    def classify(self, out):
        return self.fc(out)

    def classify1(self, out):
        return self.fc1(out)


def ResNet18(**kwargs):
    return ResNetCon(BasicBlock, [2, 2, 2, 2], **kwargs)


def ResNet34(**kwargs):
    return ResNetCon(BasicBlock, [3, 4, 6, 3], **kwargs)


def ResNet50(**kwargs):
        return ResNetCon(Bottleneck, [3, 4, 6, 3], **kwargs)


def ResNet101(**kwargs):
    return ResNetCon(Bottleneck, [3, 4, 23, 3], **kwargs)


def ResNet152(**kwargs):
    return ResNetCon(Bottleneck, [3, 8, 36, 3], **kwargs)